package org.globus.transfer.reliable.service.cache;

import java.io.IOException;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.transfer.reliable.service.MyPerformanceLog;
import org.globus.transfer.reliable.service.cache.ConnectionCache;
import org.globus.transfer.reliable.service.cache.ConnectionCacheImpl;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.ftp.exception.ServerException;
import commonj.timers.TimerManager;
import org.ietf.jgss.GSSCredential;

public class ConnectionManager
{
    private static Log logger = LogFactory.getLog(
        ConnectionManager.class.getName());
   
    private static final int NUMBER_CACHES = 3;
    public static final int SINGLE_CONNECTION = 0;
    public static final int MLSD_CONNECTION = 1;
    public static final int THIRDPARTY_CONNECTION = 2;

    // default settings
    private static final int DEFAULT_CONNECTION_TIMEOUT = 1000 * 60;
    private static final int DEFAULT_SWEEPER_INTERVAL = 1000 * 15;
    private static final int DEFAULT_MAX_OPEN_FDS = 80;
 
    private Sweeper sweeper;
    private ConnectionCounter connectionCounter;
    private ConnectionCache[] caches;
    private long connectionTimeout = -1;
    private int sweeperInterval = -1;
    private int maxNumberCachedFDs = -1;

    /**
     * Get a new instance of a connection manager
     * @param cacheConfig
     * @param timerManager
     * @return
     * @throws RftException
     */
    public static ConnectionManager createNewInstance(
        ConnectionCacheConfiguration cacheConfig,
        TimerManager                 timerManager)
    throws RftException
    {    
        ConnectionManager connectionManager = new ConnectionManager();
        connectionManager.init(cacheConfig,timerManager);
        return connectionManager;
    }
    
    /**
     * Private constructor
     */
    private ConnectionManager() {}
    
    /**
     * Add a connection to the cache
     * @param resource
     */
    public void addConnection(Connection connection)
    {
        int ret = 0;
        boolean cacheIsFull = false;
        MyPerformanceLog log1 = new MyPerformanceLog("Get biggest cache and shrink");
        MyPerformanceLog log2 = new MyPerformanceLog("Add connection");

        do
        {
            // check if cache is already full
            cacheIsFull = ((this.connectionCounter.getValue() +
                    connection.getNumberFDs()) > this.maxNumberCachedFDs);
  
            if (cacheIsFull) {
            	if (logger.isDebugEnabled()) {
                    logger.debug("shrinking cache");
                }
            	log1.start();
                ret = caches[getIndexOfCacheWithOldestConnection()].shrink();
                log1.stop();
            }
        } while (cacheIsFull && (ret != -1));

        int cacheIndex = connection.getType();

        if (logger.isDebugEnabled())
        {
            logger.debug("Type of connection: " + cacheIndex);
            if (ret == -1)
            {
                logger.debug("Cache is full -- cannot add connection.");
            }
        }

        if (ret != -1)
        {
            log2.start();
            caches[cacheIndex].addConnection(connection);
            log2.stop();
        }

        this.sweeper.scheduleSweeper();
    }
    
    /**
     * Get a connection. If no suitable connection is found
     * in the cache a new one is generated, which will be authorized with the
     * specified subjectName.
     * @param connectionType
     * @param params
     * @param credential
     * @param subjectName
     * @return
     * @throws RftException Thrown only if wrong connection type is given
     * @throws RftTransientException Thrown in any other case: interaction
     *                               with GridFTP server failed
     */
    public Connection getConnection(
           int                  connectionType,
           ConnectionParameter  params,
           GSSCredential        credential,
           String               subjectName)
    throws RftException, RftTransientException
    {
        return this.getConnection(connectionType, params, credential,
            subjectName, subjectName);
    }
    
    /**
     * Get a connection. If no suitable connection is found
     * in the cache a new one is generated which will be authorized with
     * the specified subjectNames. If the connectionType is a single or an
     * mlsd connection the destinationSubjectName will be ignored.
     * @param connectionType
     * @param params
     * @param credential
     * @param sourceSubjectName for authorization to source server
     * @param destSubjectName for authorization to destination server
     * @return
     * @throws RftException Thrown only if wrong connection type is given
     * @throws RftTransientException Thrown in any other case: interaction
     *                               with GridFTP server failed
     */
    public Connection getConnection(
           int                  connectionType,
           ConnectionParameter  params,
           GSSCredential        credential,
           String               sourceSubjectName,
           String               destSubjectName)
    throws RftException, RftTransientException
    {
        MyPerformanceLog log1 = null;
        MyPerformanceLog log2 = null;
        Connection connection = null;

        if ((connectionType < SINGLE_CONNECTION) ||
            (connectionType > THIRDPARTY_CONNECTION))
        {
            throw new RftException("invalid connection type: " + connectionType);
        }

        int cacheIndex = connectionType;

        // get a connection from the cache
        if (logger.isDebugEnabled())
        {
            log1 = new MyPerformanceLog("Check for cached connection");
     	    log1.start();
        }

        connection = caches[cacheIndex].getConnection(params);
        if (logger.isDebugEnabled())
        {
            log1.stop();
        }

        // create a new connection if there was no cached connection
        if (connection == null)
        {
            if (logger.isDebugEnabled())
            {
                logger.debug("creating new connection of type " +
                             connectionType);
     	        log2 = new MyPerformanceLog("Create connection");
                log2.start();
    	    }

            connection = createNewConnection(
                connectionType, params, credential, sourceSubjectName,
                destSubjectName);

            if (logger.isDebugEnabled())
            {
                log2.stop();
            }
        }
        else
        {
            // set the path(s) in the cached connection
    	    if (logger.isDebugEnabled())
            {
                logger.debug("found cached connection of type " +
                             connectionType);
    	    }
            connection.updateForReuse(params);
        }

        return connection;
    }
    
    /**
     * Remove expired connections in all caches.
     */
    public void removeExpiredConnections()
    {
        for(int i=0; i<this.caches.length; i++) {
            caches[i].removeExpiredConnections();
        }        
    }

    /**
     * Close all connections in all caches.
     */
    public void closeAllConnections()
    {    
        for(int i=0; i<this.caches.length; i++)
        {
            this.caches[i].closeAllConnections();
        }
    }
    
    /**
     * Get the number of cached file descriptors.
     * @return
     */
    public int getNumberCachedFDs()
    {
        return this.connectionCounter.getValue();
    }

    /**
     * Initialize the connectionManager
     * @param cacheConfig object that holds configuration parameters
     * @param timerManager TimerManager to enable periodical checks for
     *        expired connections
     * @throws RftException
     */
    private void init(
        ConnectionCacheConfiguration cacheConfig,
        TimerManager                 timerManager)
    throws RftException
    {
        if (cacheConfig == null) {
            String errorMessage = "Cache configuration null.";
            logger.error(errorMessage);
            throw new RftException(errorMessage);
        }
        if (timerManager == null) {
            String errorMessage = "TimerManager null.";
            logger.error(errorMessage);
            throw new RftException(errorMessage);
        }

        this.connectionCounter = new ConnectionCounter();
        this.maxNumberCachedFDs = cacheConfig.getMaxNumberCachedFDs();
        this.connectionTimeout = cacheConfig.getConnectionTimeout();
        this.sweeperInterval = cacheConfig.getSweeperInterval();

        // check the settings and correct them if necessary
        if (this.maxNumberCachedFDs < 0) {
            this.maxNumberCachedFDs = DEFAULT_MAX_OPEN_FDS;
            logger.warn("Cache size < 0 in RFT connection cache configuration."
                + " Setting it to default value " + this.maxNumberCachedFDs);
        }
        if (this.connectionTimeout < 0) {
            this.connectionTimeout = DEFAULT_CONNECTION_TIMEOUT;
            logger.warn("Connection timeout < 0 in RFT connection cache "
                + "configuration. Setting it to default value "
                + this.connectionTimeout);
        }
        if (this.sweeperInterval < 1000) {
            this.sweeperInterval = DEFAULT_SWEEPER_INTERVAL;
            logger.warn("Sweeper interval < 1s in RFT connection cache "
                + "configuration. Setting it to default value "
                +  this.sweeperInterval);
        }

        // create one cache for each connection type
        this.caches = new ConnectionCacheImpl[NUMBER_CACHES];
        for(int type=0; type < NUMBER_CACHES; type++)
        {
            caches[type] = new ConnectionCacheImpl(
                type, connectionTimeout, this.connectionCounter);
        }

        this.sweeper = new Sweeper(
            timerManager, this, sweeperInterval);
    }

    /**
     * Create a new connection
     * @param connectionType
     * @param rftOptions
     * @param params
     * @param credential
     * @return
     * @throws RftException Thrown only if wrong connection type is given
     * @throws RftTransientException Thrown in any other case: interaction
     *                               with GridFTP server failed
     */
    private Connection createNewConnection (
        int                  connectionType,
        ConnectionParameter  params,
        GSSCredential        credential,
        String               sourceSubjectName,
        String               destSubjectName)
    throws  RftException, RftTransientException
    {
        Connection connection = null;
        try {
            if (connectionType == SINGLE_CONNECTION) {
            	if (logger.isDebugEnabled()) {
            	    logger.debug("creating new SINGLE_CONNECTION");
            	}
            	connection = new SingleConnectionImpl(
                    params, credential, sourceSubjectName);
            } else if (connectionType == MLSD_CONNECTION) {
            	if (logger.isDebugEnabled()) {
            	    logger.debug("creating new MLSD_CONNECTION");
            	}
                connection = new MlsdConnectionImpl(
                    params, credential, sourceSubjectName);
            } else if (connectionType == THIRDPARTY_CONNECTION) {
            	if (logger.isDebugEnabled()) {
            	    logger.debug("creating new THIRDPARTY_CONNECTION");
            	}
                connection = new ThirdPartyConnectionImpl(
                    params, credential, sourceSubjectName, destSubjectName);
            } else {
                throw new RftException("invalid connection type");
            }
        } catch (IOException ioe) {
            logger.error("Can't create connection: "+ioe.toString());
            throw new RftTransientException("Connection creation error",ioe);
        } catch (ServerException se) {
            logger.error("Can't create connection: "+se.toString());
            throw new RftTransientException("Connection creation error",se);
        }
        return connection;
    }
    
    /**
     * Get the index of the cache with the connection that has not been used
     * for the longest time.
     * @return cache index
     */
    private int getIndexOfCacheWithOldestConnection()
    {
        int index = 0;
        long oldestTimestamp = Long.MAX_VALUE;
        for(int i = 0; i < NUMBER_CACHES; i++)
        {
            long timestamp = caches[i].getOldestTimestamp();
            if (timestamp > 0 && timestamp < oldestTimestamp)
            {
                oldestTimestamp = timestamp;
                index = i;
            }
        }
        return index;
    }    
}