package org.globus.transfer.reliable.service.cache;

import java.io.IOException;
import java.util.Date;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.rft.generated.RFTOptionsType;
import org.globus.transfer.reliable.service.UserSubjectUtil;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.ftp.ByteRangeList;
import org.globus.ftp.DataChannelAuthentication;
import org.globus.ftp.GridFTPClient;
import org.globus.ftp.GridFTPRestartMarker;
import org.globus.ftp.GridFTPSession;
import org.globus.ftp.MlsxEntry;
import org.globus.ftp.RetrieveOptions;
import org.globus.ftp.Session;
import org.globus.ftp.exception.ServerException;
import org.globus.ftp.vanilla.Reply;
import org.globus.gsi.gssapi.auth.IdentityAuthorization;
import org.ietf.jgss.GSSCredential;

/**
 * Class for all single connections.
 * MlsdConnectionImpl extends from it, ThirdPartyConnectionImpl connects two
 * of these and URLExpanderCleint uses a SingleConnectionImpl next to an
 * MlsdConnection for its task. 
 */
public class SingleConnectionImpl extends GridFTPClient
                                  implements Connection
{
    private static Log logger = 
        LogFactory.getLog(SingleConnectionImpl.class.getName());
    
    private static final int DEFAULT_PARALLELISM = 1;
    private static final int DEFAULT_TCP_BUFFER_SIZE = 16384;
    private static final int numberFDs = 1;
    
    protected String userDN;
    protected long lastUsed;
    protected String path;
    protected int hashCode;
    protected int type;
    protected boolean supportsRDEL;
    
    /**
     * Create a connection to a GridFTPServer, do authentication
     * and authorization
     * @param params Parameters of the connection
     * @param rftOptions Options to be set for the connection
     * @param credential for authentication at the gridftp server
     * @throws IOException If the constructor call to GridFTPClient fails.
     * @throws ServerException If the constructor call to GridFTPClient fails.
     * @throws RftTransientException If any other GridFTP server interaction
     *                               error occurs.
     */
    public SingleConnectionImpl(
           ConnectionParameter  params,
           GSSCredential        credential,
           String               authorizationSubject)
    throws IOException, ServerException, RftTransientException
    {        
        // call constructor of GridFTPClient
        super(params.getSourceHost(),params.getSourcePort());

        RFTOptionsType rftOptions = params.getRFTOptions();
        this.hashCode = params.hashCode();
        this.path = params.getSourcePath();
        this.userDN = params.getUserDN();

        // authorization
        if (!UserSubjectUtil.checkForNull(authorizationSubject)) {
            this.setAuthorization(
                new IdentityAuthorization(authorizationSubject));
        } else {
            if(logger.isDebugEnabled()) {
                logger.debug("authorizationDN not set in connection." +
                    " Using HostAuthorization.");
            }
        }
        
        // authentication
        String userName = rftOptions.getUserName();
        if (userName != null && !userName.equalsIgnoreCase("null")) {
            try {
                this.authenticate(credential, userName);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Authentication with credential and username failed on " +
                    "server " + this.getHost(), e);
            }
        } else {
            try {
                this.authenticate(credential);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Authentication with credential only failed on server " +
                    this.getHost(), e);
            }
        }

        try {
            this.setMode(GridFTPSession.MODE_EBLOCK);
        } catch (Exception e) {
            throw new RftTransientException(
                "Setting ModeE on " + this.getHost() + " failed",e);
        }
        
        // set transfer-options
        this.setOptions(rftOptions);
        
        // check if the server supports recursive deletion
        this.supportsRDEL = this.checkForRdel();
        
        this.type = ConnectionManager.SINGLE_CONNECTION;
    }

    /**
     * Make a directory if it does not exist already
     * @param dir Name of the directory
     * @throws RftTransientException If any GridFTP server interaction
     *                               error occurs.
     */
    public void makeDirectory(String dir)
    throws RftTransientException
    {
        boolean dirExists = false;
        
        // check for existence of directory
        try {
        	dirExists = super.exists(dir);
        } catch (Exception e) {
            throw new RftTransientException(
                "Check for existence of directory " + dir +
                " failed on server " + this.getHost(), e);
        }
        
        // create it if it does not exist
        if (!dirExists) {
            try {
                super.makeDir(dir);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Creating directory " + dir + 
                    " failed on server " + this.getHost(), e);
            }
        }
        updateLastUsed();
    }
    
    /**
     * Change to the given directory. If it does not exist the directory
     * will be created
     * @param dirString Path of directory
     * @param createIfNotExist If true create the directory if it does not yet
     *                         exist.
     * @throws RftTransientException If any GridFTP server interaction
     *                               error occurs.
     */
    public void changeDirectory(
           String     dirString,
           boolean    createIfNotExist)
    throws RftTransientException
    {
        try {
        	// change directory
            super.changeDir(dirString);
        } catch (Exception e) {
            if (createIfNotExist) {
            	// create directory
                try {
                    this.makeDirectory(dirString);
                } catch (RftTransientException te) {
                    throw te;
                }
                // change to newly created directory
                try {
                    super.changeDir(dirString);
                } catch (Exception e2) {
                    throw new RftTransientException("" +
                        "Changing to directory " + dirString + 
                        " failed on server " + this.getHost(), e2);
                }
            } else {
                throw new RftTransientException(
                    "Changing to directory " + dirString + 
                    " failed on server " + this.getHost(), e);
            }
        }
        updateLastUsed();
    }

    
    /**
     * Delete a file or directory
     * @throws RftException If a directory shall be removed and the GridFTP
     *                      server does not support the RDEL command
     * @throws RftTransientException If any other GridFTP server interaction
     *                               error occurs.
     */
    public void delete()
    throws RftException, RftTransientException
    {
        if (this.path.endsWith("/")) {
            if(this.supportsRDEL) {
                this.deleteDirectory();
            } else {
                throw new RftException(
                    "Server " + this.getHost() + " too old: does not " +
                    "support recursive deletion");
            }
        } else {
            try {
                super.deleteFile(this.path);
            } catch (Exception e) {
            	try {
            	    if (super.exists(this.path)) {
                        throw new RftTransientException(
                            "Deleting file " + this.path +
                            "failed on server "+ this.getHost(),e);
            	    }
            	} catch (Exception e2) {
                    throw new RftTransientException(
                        "Check for existence of " + this.path +
                        "failed on server "+ this.getHost(),e);
            	}
            }
        }
        updateLastUsed();
    }

    /**
     * Set permissions on the current path.
     * @param permissions
     * @throws RftException If the permissions are null.
     * @throws RftTransientException If the site command fails.
     */
    public void setPermissions(String permissions)
    throws RftException, RftTransientException
    {
        if (permissions == null) {
            throw new RftException("Permissions are null");
        }
        
        try {
            String command = "CHMOD " + permissions + " " + this.path;
            this.site(command);
        } catch (Exception e) {
            throw new RftTransientException(
                "Setting permissions on " + this.path +
                " failed on server " + this.getHost(), e);
        }
    }
    
    /**
     * Get permissions of the current path.
     * @return Result of the MLST command
     * @throws RftException          If the file/dir doesn't exist
     * @throws RftTransientException If the mlst call fails on an existing
     *                               file/path.
     */
    public MlsxEntry doMlst()
    throws RftException, RftTransientException
    {
        try {
            return this.mlst(this.path);
        } catch (Exception e) {
            boolean exists = false;
        	try {
                exists = super.exists(this.path);
            } catch (Exception e2) {
                throw new RftTransientException(
                    "Check for existence of " + this.path +
                    " failed on server " + this.getHost(), e2);
            }
            if (exists) {
                throw new RftTransientException(
                    "Getting file settings of " + this.path +
                    " failed on server " + this.getHost(), e);
            } else {
            	throw new RftException(
                    "Can't do MLST on non-existing file/dir " + this.path +
                    " on server " + this.getHost(),e);
            }
        }
    }
    
    /**
     * Set the restartMarker attribute of the connection object.
     * @param marker The new restartMarker value
     * @throws RftTransientException If setRestartMarker call fails.
     */
    public void setRestartMarker(String marker)
    throws RftTransientException
    {
        try {
            marker = "Range Marker " + marker;
            GridFTPRestartMarker restartmarker
                = new GridFTPRestartMarker(marker);
            ByteRangeList list = new ByteRangeList();
            list.merge(restartmarker.toVector());
            this.setRestartMarker(list);
        } catch (Exception e) {
            throw new RftTransientException(
                "Setting restart marker failed on server " +
                this.getHost(), e);
        }
        
    }
    
    /**
     * Close the connection to the GridFTP server. Ignore all exceptions.
     */
    public void closeConnection()
    {
        try {
            super.close(true);
        } catch (Exception e) {}
    }

    /**
     * Send a NOOP over the ControlChannel. If any exception occurs
     * the client should not reuse it
     * @return
     */
    public boolean checkConnectionIntegrity()
    {
        try {
            this.quote("NOOP");
            return true;
        } catch (Exception e) {
            return false;
        }
    }

    /**
     * Check if this client can be reused.
     */
    public boolean checkForReuse (ConnectionParameter params)
    {
        return (this.hashCode == params.hashCode() &&
            this.userDN.equals(params.getUserDN()));
    }
    
    public int getType()
    {
        return this.type;
    }
    
    public int getNumberFDs()
    {
        return numberFDs;
    }
    
    public long getLastUsed()
    {
        return this.lastUsed;
    }
    
    public String getPath()
    {
        return this.path;
    }
    
    public void setPath(String path)
    {
        this.path = path;
    }
        
    public void updateLastUsed()
    {
        this.lastUsed = new Date().getTime();
    }

    public void updateForReuse(ConnectionParameter params)
    {
        this.path = params.getSourcePath();
    }

    /**
     * Delete a directory.
     * @throws RftTransientException If any GridFTP server interaction
     *                               error occurs.
     */
    public void deleteDirectory()
    throws RftTransientException
    {
        boolean dirExists = false;

        // check if directory exists
        try {
            dirExists = super.exists(this.path);
        } catch (Exception e) {
            throw new RftTransientException(
                "Check for existence of directory " + this.path +
                " failed on server " + this.getHost(), e);
        }
        
        // Delete directory if it exists
        if (dirExists) {
            // Try to send SITE rdel <filepath> 
            String recursiveDeleteArgs = "rdel " + this.path; 

            try {
                super.site(recursiveDeleteArgs);
            } catch (Exception e) {
                throw new RftTransientException(
                      "Recursive directory removal of " + this.path +
                      " failed on server " + this.getHost(), e);
            }
        }
    }

    /**
     * Set options for the transfer
     * @param rftOptions Options to be set for the connection
     * @throws RftTransientException If any GridFTP server interaction
     *                               error occurs.
     */
    private void setOptions(RFTOptionsType rftOptions)
    throws RftTransientException
    {
        int parallelism = 0;
        int tcpBufferSize = 0;
        boolean binaryTransfer = true;
        boolean dcau = false;
        
        if (rftOptions != null) {
            if (rftOptions.getParallelStreams() != null) {
                parallelism = rftOptions.getParallelStreams().intValue();
            }
            if (rftOptions.getTcpBufferSize() != null) {
                tcpBufferSize = rftOptions.getTcpBufferSize().intValue();
            }
            if (rftOptions.getBinary() != null) {
                binaryTransfer = rftOptions.getBinary().booleanValue();
            }
            if (rftOptions.getDcau() != null) {
                dcau = rftOptions.getDcau().booleanValue();
            }
        }

        // data channel authentication
        if (dcau) {
            // set data channel authentication to SELF
            try {
                this.setDataChannelAuthentication(
                    DataChannelAuthentication.SELF);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Setting data channel autentication to SELF failed " +
                    "on server " + this.getHost(), e);
            }
            
            // set protection buffer size
            try {
                this.setProtectionBufferSize(16384);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Setting protection buffer size failed on server " +
                    this.getHost(), e);
            }
        } else {
            // set data channel authentication to NONE
            try {
                this.setDataChannelAuthentication(
                    DataChannelAuthentication.NONE);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Setting data channel autentication to NONE failed " +
                    "on server " + this.getHost(),e);
            }
        }

        // transfer type
        try {
            if (binaryTransfer) {
                super.setType(Session.TYPE_IMAGE);
            } else {
                super.setType(Session.TYPE_ASCII);
            }
        } catch (Exception e) {
            throw new RftTransientException(
                "Setting transfer type failed on server " +
                this.getHost(), e);
        }

        // TODO: set modeE here

        // parallelism
        try {
            if (parallelism == 0) {
                parallelism = DEFAULT_PARALLELISM;
            }
            super.setOptions(new RetrieveOptions(parallelism));
        } catch (Exception e) {
            throw new RftTransientException(
                "Setting parallelism failed on server " +
                this.getHost(), e);
        }
        
        // TCP buffer size
        try {
            if (tcpBufferSize == 0) {
                tcpBufferSize = DEFAULT_TCP_BUFFER_SIZE;
            }
            super.setTCPBufferSize(tcpBufferSize);
        } catch (Exception e) {
            throw new RftTransientException(
                "Setting TCP buffersize failed on server " +
                this.getHost(), e);
        }
        
    }

    /**
     * Check if the GridFTPServer supports the recursive deletion.
     * @throws RftException In case the GridFTP server doesn't support RDEL.
     * @throws RftTransientException In case the check for RDEL itself fails.
     * @return true if server supports RDEL
     */
    private boolean checkForRdel()
    throws RftTransientException
    {
        Reply reply;
        try {
            reply = this.site("HELP SITE");
        } catch (Exception e) {
             throw new RftTransientException(
                 "HELP SITE command failed on server " +
                 this.getHost() ,e);
        }
        
        if (-1 == reply.toString().indexOf("SITE RDEL")) {
            return false;
        } else {
            return true;
        }
    }

}
