package org.globus.transfer.reliable.service.cache;

import java.util.Date;
import java.io.IOException;
import org.globus.ftp.MlsxEntry;
import org.globus.ftp.HostPort;
import org.globus.ftp.HostPortList;
import org.globus.ftp.exception.ServerException;
import org.globus.transfer.reliable.service.MyMarkerListener;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.transfer.reliable.service.exception.RftException;
import org.ietf.jgss.GSSCredential;

/**
 * Class for a third party connection
 */
public class ThirdPartyConnectionImpl implements Connection
{
    protected SingleConnectionImpl source;
    protected SingleConnectionImpl destination;
    protected String userDN;
    protected long lastUsed;
    protected int hashCode;
    protected int type;
    private static final int numberFDs = 2;
    
    public ThirdPartyConnectionImpl (
           ConnectionParameter  params,
           GSSCredential        credential,
           String               sourceSubjectName,
           String               destSubjectName)
    throws IOException, ServerException, RftTransientException, RftException
    {
        this.userDN = params.getUserDN();

        // create connection to source GridFTP server
        source = new SingleConnectionImpl(
                     new ConnectionParameter(
                         params.getSourceHost(),
                         params.getSourcePort(),
                         params.getSourcePath(),
                         params.getUserDN(),
                         params.getRFTOptions()),
                     credential,
                     sourceSubjectName);
         
        // create connection to destination GridFTP server
        destination = new SingleConnectionImpl(
                          new ConnectionParameter(
                              params.getDestinationHost(),
                              params.getDestinationPort(),
                              params.getDestinationPath(),
                              params.getUserDN(),
                              params.getRFTOptions()),
                          credential,
                          destSubjectName);

        boolean sourceIsStripedEnabled;
        boolean destIsStripedEnabled;
        
        // Check for striping support on source
        try {
        	sourceIsStripedEnabled = source.isFeatureSupported("SPOR");
        } catch (Exception e) {
            throw new RftTransientException(
                "Check for Striping on source " + source.getHost() +
                " failed",e);
        }

        // Check for striping support on destination
        try {
            destIsStripedEnabled = destination.isFeatureSupported("SPAS");
        } catch (Exception e) {
            throw new RftTransientException(
                "Check for Striping on destination " + destination.getHost() +
                " failed",e);
        }

        // activate striping if both servers support it
        if (sourceIsStripedEnabled && destIsStripedEnabled) {
            HostPortList hostPortList;
        	// set destination to striped passive
        	try {
                hostPortList = destination.setStripedPassive();
            } catch (Exception e) {
                throw new RftTransientException(
                    "Setting destination " + destination.getHost() + 
                    " to striped passive failed",e);
                
            }
            // set source to striped active
            try {
                source.setStripedActive(hostPortList);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Setting source " + source.getHost() +
                    " to striped active failed",e);
            }
        } else {
            HostPort hostPort;
            // set destination to passive
        	try {
            	hostPort = destination.setPassive();
        	} catch (Exception e) {
                throw new RftTransientException(
                    "Setting destination " + destination.getHost() + 
                    " to passive failed",e);
        	}
        	// set source to active
        	try {
                source.setActive(hostPort);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Setting source " + source.getHost() + 
                    " to active failed",e);
            }
        }

        this.hashCode = params.hashCode();
        this.type = ConnectionManager.THIRDPARTY_CONNECTION;
    }
        
    /**
     * Set the restartMarker attribute of the TransferClient object.
     * @param marker The new restartMarker value
     * @throws RftTransientException If setting restart marker on the
     *                               server fails
     */
    public void setRestartMarker(String marker)
    throws RftTransientException
    {
        source.setRestartMarker(marker);
    }
    
    /**
     * Close the third party connection
     */
    public void closeConnection()
    {
        source.closeConnection();
        destination.closeConnection();
    }
    
    /**
     * Check if the existing connection can be reused for another transfer.
     * Sanity security check: compare the subjects, just in case the hashCode
     * creation code has problems.
     * @param transferJob
     * @param rftOptions
     * @return true in case the connection can be reused
     */
    public boolean checkForReuse (
           ConnectionParameter params)
    {    
        return (this.hashCode == params.hashCode() &&
            this.userDN.equals(params.getUserDN()));
    }
    
    /**
     * Check if 3rd party connection is still alive and ok
     * @return true if the connection is ok
     */
    public boolean checkConnectionIntegrity()
    {
        return (   source.checkConnectionIntegrity()
                && destination.checkConnectionIntegrity());
    }

    public void updateForReuse(ConnectionParameter params)
    {
        this.source.setPath(params.getSourcePath());
        this.destination.setPath(params.getDestinationPath());
    }
        
    /**
     * Perform an extended 3rd party transfer
     * @throws RftTransientException
     */
    public void normalExtendedTransfer(MyMarkerListener markerListener)
    throws RftTransientException {
        try {
            this.source.extendedTransfer(
                    this.source.getPath(),
                    this.destination,
                    this.destination.getPath(),
                    markerListener);
        } catch (Exception e) {
            throw new RftTransientException(
                "Extended 3rd-party transfer " + 
                this.source.getHost() + ":" + this.source.getPort() +
                this.source.getPath() + " --> " + this.destination.getHost() +
                ":" + this.destination.getPort() +
                this.destination.getPath() + " failed", e);
        }
    }

    /**
     * Perform a non-extended 3rd party transfer
     * @throws RftTransientException
     */
    public void normalNonExtendedTransfer(MyMarkerListener markerListener) 
    throws RftTransientException 
    {
        try {
            this.source.transfer(
                    this.source.getPath(),
                    this.destination,
                    this.destination.getPath(),
                    false,
                    markerListener);
        } catch (Exception e) {
            throw new RftTransientException(
                "Non-Extended 3rd-party transfer " + 
                this.source.getHost() + ":" + this.source.getPort() +
                this.source.getPath() + " --> " + this.destination.getHost() +
                ":" + this.destination.getPort() +
                this.destination.getPath() + " failed", e);
        }
    }
    
    /**
     *
     * @return
     * @throws RftException If the file/dir doesn't exist
     * @throws RftTransientException If the mlst call fails on an existing
     */
    public MlsxEntry doMlstOnSource()
    throws RftException, RftTransientException
    {
        return this.source.doMlst();
    }
    
    /**
     * Set permissions of files on the destination
     * @param permissions
     * @throws RftException If the permissions are null
     * @throws RftTransientException If setting permissions fails
     */
    public void setPermissionsOnDestination(String permissions)
    throws RftException, RftTransientException
    {
        this.destination.setPermissions(permissions);
    }
    
    public int getType()
    {
        return this.type;
    }
    
    public int getNumberFDs()
    {
        return numberFDs;
    }
    
    public long getLastUsed()
    {
        return this.lastUsed;
    }
    
    public String getSourcePath()
    {
        return this.source.getPath();
    }
    
    public String getDestinationPath()
    {
        return this.destination.getPath();
    }
    
    public void updateLastUsed()
    {
        this.lastUsed = new Date().getTime();
    }

}
