package org.globus.transfer.reliable.service.client;

import java.io.File;
import java.io.IOException;
import java.io.FileNotFoundException;
import org.globus.util.Util;
import org.globus.wsrf.ResourceProperty;
import org.globus.wsrf.ResourcePropertySet;
import org.globus.ftp.DataSink;
import org.globus.ftp.DataSource;
import org.globus.ftp.FileRandomIO;
import org.globus.ftp.MlsxEntry;
import org.globus.rft.generated.RFTOptionsType;
import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.globus.transfer.reliable.service.cache.ConnectionManager;
import org.globus.transfer.reliable.service.cache.ConnectionParameter;
import org.globus.transfer.reliable.service.cache.SingleConnectionImpl;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.transfer.reliable.service.MyMarkerListener;
import org.globus.transfer.reliable.service.RFTConstants;
import org.globus.transfer.reliable.service.UserSubjectUtil;
import org.globus.rft.generated.TotalBytesTransferredResource;
import org.ietf.jgss.GSSCredential;

public class GetPutTransferClient implements Client
{
    private long size;
    private int transferId;
    private String restartMarker;
    private String permissions;
    private ConnectionManager connectionManager = null;
    private SingleConnectionImpl sourceConnection = null;
    private SingleConnectionImpl destinationConnection = null;
    private ResourceProperty totalBytesTransferredRP;
    private TotalBytesTransferredResource totalBytes;
    
    // TODO: Is the totalBytesTransferred (rpSet) ok?
    // TODO: what to do with extended transfers?
    
    public GetPutTransferClient(
        ConnectionParameter  sourceParams,
        ConnectionParameter  destinationParams,
        GSSCredential        credential,
        ConnectionManager    connectionManager,
        ResourcePropertySet  rpSet,
        int                  transferId)
    throws RftException, RftTransientException, RftDBException
    {
        // check parameter
    	if (sourceParams == null) {
            throw new RftException("parameter for source connection " +
                "must not be null");
        }
        if (destinationParams == null) {
            throw new RftException("parameter for destination connection " +
                "must not be null");
        }
        if (credential == null) {
        	throw new RftException("credential must not be null");
        }
        if (connectionManager == null) {
            throw new RftException("connectionManager must not be null");
        }
        if (rpSet == null) {
            throw new RftException("ResourcePropertySet " +
                "must not be null");
        }
        
        RFTOptionsType rftOptions = sourceParams.getRFTOptions();
        this.connectionManager = connectionManager;
        this.sourceConnection
            = (SingleConnectionImpl)
                  this.connectionManager.getConnection(
                      ConnectionManager.SINGLE_CONNECTION,
                      sourceParams,
                      credential,
                      UserSubjectUtil.getSourceUserSubject(rftOptions));
        
        this.destinationConnection
            = (SingleConnectionImpl)
                  this.connectionManager.getConnection(
                      ConnectionManager.SINGLE_CONNECTION,
                      destinationParams,
                      credential,
                      UserSubjectUtil.getDestinationUserSubject(rftOptions));
        
        this.transferId = transferId;
        this.totalBytesTransferredRP
            = rpSet.get(RFTConstants.TOTAL_BYTES_RESOURCE);
        this.totalBytes
            = (TotalBytesTransferredResource) totalBytesTransferredRP.get(0);

        this.restartMarker = ReliableFileTransferDbAdapter
                                 .getDbAdapter()
                                     .getRestartMarker(transferId);
    }
    
    public void process()
    throws RftException, RftTransientException, RftDBException
    {
        // get permissions and size from the source
        this.doMlstOnSource();
        
        // do the transfer
        this.transfer();
        
        // set permissions on the destination
        destinationConnection.setPermissions(permissions);
        
        // update total bytes
        this.totalBytes.setTotalBytes(
                this.totalBytes.getTotalBytes() + this.size);
        this.totalBytesTransferredRP.set(0, this.totalBytes);
        
        // update last used and add connection back to the cache
        this.sourceConnection.updateLastUsed();
        this.destinationConnection.updateLastUsed();
        this.connectionManager.addConnection(sourceConnection);
        this.connectionManager.addConnection(destinationConnection);
    }

    /**
     * Perform a transfer from source to destination in a non-
     * third-party manner: get the file from the source, store it
     * in a temporary file an put it to the destination. Depending
     * on the configuration these transfers are done in extended mode
     * or not.
     * @throws RftException
     * @throws RftTransientException
     */
    private void transfer()
    throws RftException, RftTransientException
    {
    	MyMarkerListener markerListener = new MyMarkerListener(transferId);
        File fullLocalFile;
    	try {
            fullLocalFile = File.createTempFile("TempRFT", 
                String.valueOf(this.transferId));
    	} catch (IOException e) {
    	    throw new RftException(
    	        "Failed to create tmp file", e);
    	}
       
    	Util.setOwnerAccessOnly(fullLocalFile.getAbsolutePath());
        fullLocalFile.deleteOnExit();
        
        DataSink sink;
        try {
            sink = new FileRandomIO(
                       new java.io.RandomAccessFile(
                           fullLocalFile,"rw"));
        } catch (FileNotFoundException e) {
    	    throw new RftException(
        	    "Failed to create data sink", e);            
        }
        
        // get file from source, choose extended mode if the
        // client is configured to do so
        if (this.restartMarker != null) {
            this.sourceConnection.setRestartMarker(this.restartMarker);
            try {
                this.sourceConnection.extendedGet(
                    sourceConnection.getPath(),
                    this.size,
                    sink,
                    markerListener);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Extended get of " + sourceConnection.getPath() +
                    " failed on server " + sourceConnection.getHost(), e);
            }
        } else {
            try {
                this.sourceConnection.get(
                    sourceConnection.getPath(),
                    sink,
                    markerListener);
            } catch (Exception e) {
                throw new RftTransientException(
                    "Get of " + sourceConnection.getPath() +
                    " failed on server " + sourceConnection.getHost(), e);
            }
        }
        
        DataSource source;
        try {
            source = new FileRandomIO(
                         new java.io.RandomAccessFile(
                             fullLocalFile, "r"));
        } catch (FileNotFoundException e) {
    	    throw new RftException(
        	    "Failed to create data source", e);            
            
        }
        
        // put the file to the server; choose extended mode if the
        // client is configured to do so
        // TODO: is this correct? what about a destination marker?
        if (this.restartMarker != null) {
            try {
        	    this.destinationConnection.extendedPut(
                    destinationConnection.getPath(),
                    source,
                    markerListener);
            } catch(Exception e) {
                throw new RftTransientException(
                    "Extended put of " + destinationConnection.getPath() +
                    " failed on server " +  destinationConnection.getHost(),
                    e);
            }
         } else {
        	 try {
                 this.destinationConnection.put(
                     destinationConnection.getPath(),
                     source,
                     markerListener);
        	 } catch (Exception e) {
                 throw new RftTransientException(
                     "Put of " + destinationConnection.getPath() +
                     " failed on server " +  destinationConnection.getHost(),
                     e);
        	 }
         }
            
         if(fullLocalFile.exists()) {
             fullLocalFile.delete();
         }
    }
    
    /**
     * Get file settings on the source side
     * @throws RftDBException
     * @throws RftException
     * @throws RftTransientException
     */
    private void doMlstOnSource()
    throws RftException, RftTransientException, RftDBException
    {
        
        MlsxEntry entry = sourceConnection.doMlst();
        this.size = Long.parseLong(entry.get(MlsxEntry.SIZE));
        this.permissions = entry.get("unix.mode");
        
        ReliableFileTransferDbAdapter
            .getDbAdapter()
                .updateTransfer(
                    this.transferId,
                    this.size,
                    this.permissions);
    }
}
