package org.globus.transfer.reliable.service.client;

import java.util.Vector;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.globus.ftp.MlsxEntry;
import org.globus.rft.generated.RFTOptionsType;
import org.globus.rft.generated.TransferType;
import org.globus.rft.generated.TransferStatusTypeEnumeration;
import org.globus.transfer.reliable.service.UserSubjectUtil;
import org.globus.transfer.reliable.service.cache.SingleConnectionImpl;
import org.globus.transfer.reliable.service.cache.MlsdConnectionImpl;
import org.globus.transfer.reliable.service.cache.ConnectionManager;
import org.globus.transfer.reliable.service.cache.ConnectionParameter;
import org.globus.transfer.reliable.service.database.ReliableFileTransferDbAdapter;
import org.globus.transfer.reliable.service.exception.RftException;
import org.globus.transfer.reliable.service.exception.RftTransientException;
import org.globus.transfer.reliable.service.database.RftDBException;
import org.ietf.jgss.GSSCredential;

public class URLExpanderClient implements Client
{   
    private static Log logger =
        LogFactory.getLog(URLExpanderClient.class.getName());

    private int requestId;
    private int numberOfTransfers;
    private RFTOptionsType rftOptions;

    private ConnectionManager connectionManager = null;
    private MlsdConnectionImpl sourceConnection = null;
    private SingleConnectionImpl destinationConnection = null;
    
    public URLExpanderClient(
           ConnectionParameter  sourceParams,
           ConnectionParameter  destinationParams,
           GSSCredential        credential,
           ConnectionManager    connectionManager,
           int                  requestId)
    throws RftException, RftTransientException
    {
    	if (sourceParams == null) {
            throw new RftException("parameter for source connection " +
                "must not be null");
        }
        if (destinationParams == null) {
            throw new RftException("parameter for destination connection " +
                "must not be null");
        }
        if (credential == null) {
        	throw new RftException("credential must not be null");
        }
        if (connectionManager == null) {
            throw new RftException("connectionManager must not be null");
        }

        this.rftOptions = sourceParams.getRFTOptions();
        this.connectionManager = connectionManager;
        
        // TODO: IMPORTANT: MUST NOT HAVE PARALLELISM > 1 OR STRIPING !!!!!
        this.sourceConnection = (MlsdConnectionImpl)
            this.connectionManager.getConnection(
                ConnectionManager.MLSD_CONNECTION,
                sourceParams,
                credential,
                UserSubjectUtil.getSourceUserSubject(rftOptions));

        this.destinationConnection = (SingleConnectionImpl)
            this.connectionManager.getConnection(
                ConnectionManager.SINGLE_CONNECTION,
                destinationParams,
                credential,
                UserSubjectUtil.getDestinationUserSubject(rftOptions));

        this.requestId = requestId;
        this.numberOfTransfers = 0;
    }
    
    /**
     * Performs a directory expansion.
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    public void process()
    throws RftException, RftTransientException, RftDBException
    {
        expandDirectory(
            sourceConnection.getPath(),
            destinationConnection.getPath());

        sourceConnection.updateLastUsed();
        destinationConnection.updateLastUsed();
        
        // return connections to the cache
        this.connectionManager.addConnection(sourceConnection);
        this.connectionManager.addConnection(destinationConnection);
    }

    public int getNumberOfTransfers()
    {
        return this.numberOfTransfers;
    }
    
    /**
     * Expand a directory: Create destination directory if if does
     * not exist and do a listing of the source directory. In case
     * there are subdirectories: do a recursive call.
     * In case there are files: store them in the database for future
     * transfer after the expanding has finished.
     * The directory names are always absolute paths.
     * @param sourceDir
     * @param destDir
     * @throws RftException
     * @throws RftTransientException
     * @throws RftDBException
     */
    private void expandDirectory(
           String   sourceDir,
           String   destDir)
    throws RftException, RftTransientException, RftDBException
    {    
    	logger.debug("Entering expandDirectory");
    	destinationConnection.makeDirectory(destDir);

        Vector v = sourceConnection.doMlsd(sourceDir);
        
        while (!v.isEmpty()) {
              MlsxEntry entry = (MlsxEntry) v.remove(0);
              String fileName = entry.getFileName();
            if (entry.get(MlsxEntry.TYPE).equals(MlsxEntry.TYPE_DIR)) {
            	String newSourceDir = sourceDir + fileName + "/";
                String newDestDir = destDir + fileName + "/";
                this.expandDirectory(newSourceDir,newDestDir);
            } else if (entry.get(MlsxEntry.TYPE).equals(
                       MlsxEntry.TYPE_FILE)) {
                String newSourceUrl = "gsiftp://" +
                    sourceConnection.getHost() + ":" +
                    sourceConnection.getPort() + sourceDir + fileName;
                String newDestUrl = "gsiftp://" + 
                    destinationConnection.getHost() + ":" + 
                    destinationConnection.getPort() + destDir + fileName;
                TransferType transferType = new TransferType();
                transferType.setSourceUrl(newSourceUrl);
                transferType.setDestinationUrl(newDestUrl);
                transferType.setAttempts(new Integer(0));
                transferType.setRftOptions(this.rftOptions);
                transferType.setStatus(
                    TransferStatusTypeEnumeration.Pending);
                // store transfer in database
                ReliableFileTransferDbAdapter
                    .getDbAdapter()
                        .storeTransferJob(requestId, transferType);
                this.numberOfTransfers++;
            }
        }
    }
    
}
